import json

from yapy.compare.output.formatter import OutputFormatter


class JSONFormatter(OutputFormatter):

    def __init__(self, plagiarism_compare):
        super().__init__(plagiarism_compare)

    def __row_result(self, row_id):
        result = []

        length = len(self.plagiarism_compare.result)

        for i in range(0, length):
            id = self.headers[i]

            result.append({
                'id': id,
                'score': self.plagiarism_compare.result[row_id][i] or self.plagiarism_compare.result[i][row_id]
            })

        return result

    def format(self):
        result = []

        for i, id in enumerate(self.headers):
            result.append({
                'id': id,
                'results': self.__row_result(i)
            })

        return json.dumps(result, indent=4)



from abc import ABC, abstractmethod


class OutputFormatter(ABC):
    def __init__(self, plagiarism_compare):
        self.plagiarism_compare = plagiarism_compare

        self.headers = list(self.plagiarism_compare.contents.keys())

    @abstractmethod
    def format(self):
        pass

    def export(self, path):
        file = open(path, "w")
        file.write(self.format())



class ParserResult:
    def __init__(self, lark_result):
        self.lark_result = lark_result
        self.tokens = self.__tokens()
        self.tokens_table = self.__tokens_table()
        self.register_number = self.__register_number()

    def __tokens(self):
        tokens = []

        for token in self.lark_result.iter_subtrees():
            token_name = token.data
            if token_name != 'program' and token_name != 'ret' and token_name != 'header':
                tokens.append(token_name)

        return tokens

    def __tokens_table(self):
        tokens = {}

        for token in self.tokens:
            if token in tokens:
                tokens[token] += 1
            else:
                tokens[token] = 1

        return tokens

    def __register_number(self):
        number = 0

        # for tree in self.lark_result.iter_subtrees():
        #    print(tree)

        return number



def sorensen_dice_coefficient(lhs, rhs):
    similarity_score = 0
    max_score = 0

    for element in lhs:
        max_score += lhs[element]
        if element in rhs:
            similarity_score += min(lhs[element], rhs[element])

    for element in rhs:
        max_score += rhs[element]

    if max_score == 0:
        return 0

    return similarity_score * 2 / max_score



import os
from yapy.compare.output.json_formatter import JSONFormatter
from yapy.parser.parser import Parser
from yapy.compare.output.html_formatter import HtmlFormatter


class PlagiarismCompare:

    def __init__(self, files=None, path=None, contents=None, threshold=None):
        self.files = files
        self.path = path
        self.contents = contents
        self.threshold = threshold

        if contents == None:
            self.__read_files()

        self.__parse_files()
        self.html_formatter = HtmlFormatter(self)
        self.json_formatter = JSONFormatter(self)

    def __read_files(self):
        if self.path != None:
            self.files = [os.path.join(self.path, file) for file in os.listdir(
                self.path) if os.path.isfile(os.path.join(self.path, file))]

        self.contents = {}

        for filepath in self.files:
            try:
                basename = os.path.basename(filepath)
                self.contents[basename] = open(
                    filepath, 'r', encoding='utf8', errors='replace').read()
            except:
                print(f'Could not read file from path {filepath}')

    def __parse_files(self):
        self.parsed_files = {}

        for key, value in self.contents.items():
            try:
                self.parsed_files[key] = Parser.parse(value)
            except Exception as e:
                print(e)
                self.parsed_files[key] = None
                print(f'Could not parse file {key}')

    def compare(self, score_function):
        self.result = [[None]*len(self.contents)
                       for ignored in range(len(self.contents))]

        results = list(self.parsed_files.values())
        for i in range(0, len(self.contents) - 1):
            for j in range(i + 1, len(self.contents)):
                lhs = results[i]
                rhs = results[j]

                if lhs is None or rhs is None:
                    score = None
                else:
                    tmp_score = score_function(
                        lhs.tokens_table, rhs.tokens_table)

                    print(self.threshold is not None and tmp_score >= self.threshold)
                    score = None if self.threshold is not None and tmp_score < self.threshold else tmp_score

                self.result[i][j] = score



from yapy.compare.output.formatter import OutputFormatter


class HtmlFormatter(OutputFormatter):

    def __init__(self, plagiarism_compare):
        super().__init__(plagiarism_compare)

    def __draw_header(self):
        headers = self.headers

        header = '<th>Results</th>'

        for element in headers:
            header += f'<th>{element}</th>'

        return f'<tr>{header}</tr>'

    def __draw_row(self, row_id):
        row_data = [self.plagiarism_compare.result[row_id][column]
                    for column in range(0, len(self.plagiarism_compare.result))]

        row = ''

        for element in row_data:
            row += f'<td>{element}</td>'

        return f'<tr><th>{self.headers[row_id]}</th>{row}</tr>'

    def format(self):
        table = self.__draw_header()

        for row in range(0, len(self.plagiarism_compare.result)):
            table += self.__draw_row(row)

        return f'<table>{table}</table>'



from setuptools import setup, find_packages

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name="yapy-arm64",
    version="0.1.2",
    author="Luis Tavares",
    python_requires='>=3.6',
    long_description_content_type='text/markdown',
    long_description=long_description,
    install_requires=required,
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
)


from lark import Lark
from yapy.parser.parser_result import ParserResult

arm64_parser = Lark(r"""
    program: header* (label | instruction)*

    label: CNAME ":"

    header: ".text"
            | ".global" /.+/
            | ".type" /.+/


    int_constant: "#"? "-"? INT | "0x"i HEXDIGIT+

    char: /"."/ | /'.'/

    w_register: "w"i INT
    x_register: "x"i INT

    ?normal_register: w_register | x_register

    ?op: normal_register | int_constant | char | "sp"i

    address: "[" (x_register | CNAME) ("," op)? "]" "!"? ("," op)?

    ?cc: "lo"i
        | "hi"i
        | "ls"i
        | "hs"i
        | "lt"i
        | "gt"i
        | "le"i
        | "ge"i
        | "eq"i
        | "ne"i
        | "mi"i
        | "pl"i
        | "vs"i
        | "vc"i
        | "cs"i
        | "cc"i

    ?instruction: arithmetic_instruction
                | bitwise_logical_instruction
                | bitfield_instruction
                | bit_byte_instruction
                | load_store_instruction
                | branch_instructions
                | conditional_instructions
                | compare_instructions

    ?arithmetic_instruction: add
                            | sub
                            | add
                            | sub
                            | neg
                            | neg
                            | mul
                            | umull
                            | umulh
                            | smull
                            | smulh
                            | madd
                            | msub
                            | mneg
                            | umaddl
                            | umsubl
                            | umnegl
                            | smaddl
                            | smsubl
                            | smnegl
                            | udiv
                            | sdiv
    
    ?bitwise_logical_instruction: and
                                | bic
                                | orr
                                | orn
                                | eor
                                | eon
                                | lsl
                                | lsr
                                | asr
                                | ror
                                | mov
                                | mvn
                                | tst

    ?bitfield_instruction: bfi
                            | ubfx
                            | sbfx
    
    ?bit_byte_instruction: cls
                            | clz
                            | rbit
                            | rev
                            | rev16
                            | rev32
                        
    ?load_store_instruction: str
                            | strb
                            | strh
                            | stur
                            | stp
                            | ldr
                            | ldrb
                            | ldrsb
                            | ldrh
                            | ldrsh
                            | ldrsw
                            | ldur
                            | ldp

    ?branch_instructions: b
                        | bl
                        | ret
                        | bcc
                        | cbz
                        | cbnz
    
    ?conditional_instructions: csel
                                | csinc
                                | csneg
                                | csinv
                                | cset
                                | csetm
                                | cinc
                                | cneg
                                | cinv

    ?compare_instructions: cmp
                            | cmn
                            
    add: ("add"i | "adds"i) normal_register "," normal_register "," op
    sub: ("sub"i | "subs"i) normal_register "," normal_register "," op
    neg: ("neg"i | "negs"i) normal_register "," op
    ngc: ("ngc"i | "ngc"i) normal_register "," normal_register
    mul: "mul"i normal_register "," normal_register "," normal_register
    umull: "umull"i x_register "," w_register "," w_register
    umulh: "umulh"i x_register "," x_register "," x_register
    smull: "smull"i x_register "," w_register "," w_register
    smulh: "smulh"i x_register "," x_register "," x_register
    madd: "madd"i normal_register "," normal_register "," normal_register "," normal_register
    msub: "msub"i normal_register "," normal_register "," normal_register "," normal_register
    mneg: "mneg"i normal_register "," normal_register "," normal_register
    umaddl: "umaddl"i x_register "," w_register "," w_register "," x_register
    umsubl: "umsubl"i x_register "," w_register "," w_register "," x_register
    umnegl: "umnegl"i x_register "," w_register "," w_register
    smaddl: "smaddl"i x_register "," w_register "," w_register "," x_register
    smsubl: "smsubl"i x_register "," w_register "," w_register "," x_register
    smnegl: "smnegl"i x_register "," w_register "," w_register
    udiv: "udiv"i normal_register "," normal_register "," normal_register
    sdiv: "sdiv"i normal_register "," normal_register "," normal_register

    and: ("and"i | "ands"i) normal_register "," normal_register "," op
    bic: ("bic"i | "bics"i) normal_register "," normal_register "," op
    orr: "orr"i normal_register "," normal_register "," op
    orn: "orn"i normal_register "," normal_register "," op
    eor: "eor"i normal_register "," normal_register "," op
    eon: "eon"i normal_register "," normal_register "," op
    lsl: "lsl"i normal_register "," normal_register "," op
    lsr: "lsr"i normal_register "," normal_register "," normal_register
    asr: "asr"i normal_register "," normal_register "," op
    ror: "ror"i normal_register "," normal_register "," op
    mov: "mov"i normal_register "," op
    mvn: "mvn"i normal_register "," op
    tst: "tst"i normal_register "," op

    bfi: "bfi"i normal_register "," normal_register "," int_constant "," int_constant
    ubfx: "ubfx"i normal_register "," normal_register "," int_constant "," int_constant
    sbfx: "sbfx"i normal_register "," normal_register "," int_constant "," int_constant

    cls: "cls"i normal_register "," normal_register
    clz: "clz"i normal_register "," normal_register
    rbit: "rbit"i normal_register "," normal_register
    rev: "rev"i normal_register "," normal_register
    rev16: "rev16"i normal_register "," normal_register
    rev32: "rev32"i x_register "," x_register

    str: "str"i normal_register "," address
    strb: "strb"i w_register "," address
    strh: "strh"i w_register "," address
    stur: "stur"i normal_register "," address
    stp: "stp"i normal_register "," normal_register "," address
    ldr: "ldr"i normal_register "," address
    ldrb: "ldrb"i w_register "," address
    ldrsb: "ldrsb"i normal_register "," address
    ldrh: "ldrh"i w_register "," address
    ldrsh: "ldrsh"i normal_register "," address
    ldrsw: "ldrsw"i x_register "," address
    ldur: "ldur"i normal_register "," address
    ldp: "ldp"i normal_register "," normal_register "," address

    b: "b"i CNAME
    bl: "bl"i CNAME
    ret: "ret"i x_register?
    bcc: ("b."i | "b"i) cc CNAME
    cbz: "cbz"i normal_register "," CNAME
    cbnz: "cbnz"i normal_register "," CNAME

    csel: "csel"i normal_register "," normal_register "," normal_register "," cc
    csinc: "csinc"i normal_register "," normal_register "," normal_register "," cc
    csneg: "csneg"i normal_register "," normal_register "," normal_register "," cc
    csinv: "csinv"i normal_register "," normal_register "," normal_register "," cc
    cset: "cset"i normal_register "," cc
    csetm: "csetm"i normal_register "," cc
    cinc: "cinc"i normal_register "," normal_register "," cc
    cneg: "cneg"i normal_register "," normal_register "," cc
    cinv: "cinv"i normal_register "," normal_register "," cc

    cmp: "cmp"i normal_register "," op
    cmn: "cmn"i normal_register "," normal_register

    %import common.WORD
    %import common.CNAME
    %import common.HEXDIGIT
    %import common.INT
    %import common.WS
    %import common.CPP_COMMENT
    %import common.C_COMMENT

    %ignore CPP_COMMENT
    %ignore C_COMMENT
    %ignore WS
    """, start="program")


class Parser:

    @staticmethod
    def parse(content):
        return ParserResult(arm64_parser.parse(content))
